package dev.pfaff.altargui.pathfinder.fontkit;

import dev.pfaff.altargui.libimage.ColorType;
import dev.pfaff.altargui.libimage.ImageData;
import dev.pfaff.altargui.text.Font;
import dev.pfaff.altargui.text.GlyphRef;
import dev.pfaff.altargui.texture.ImageDataLoadableTexture;
import dev.pfaff.altargui.texture.LoadableTexture;
import dev.pfaff.altargui.util.GLThread;
import it.unimi.dsi.fastutil.ints.Int2ObjectMap;
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap;

import java.lang.foreign.Arena;
import java.util.Objects;
import java.util.function.IntFunction;

import static dev.pfaff.altargui.pathfinder.fontkit.FKGlyphLoadableTexture.*;
import static dev.pfaff.altargui.util.BasicFuture.executeOn;
import static java.lang.foreign.ValueLayout.JAVA_INT;

public final class FKFont implements Font {
	private static final int GLYPH_SIZE = 64;
	private static final int PAGE_DIM_GLYPHS = 16;
	private static final int PAGE_DIM_SIZE = PAGE_DIM_GLYPHS * GLYPH_SIZE;
	private static final int PAGE_GLYPHS = PAGE_DIM_GLYPHS * PAGE_DIM_GLYPHS;
	private static final int PAGE_SIZE = PAGE_DIM_SIZE * PAGE_DIM_SIZE;

	private final FKFontRef fkFontRef;
	private final Int2ObjectMap<Page> pages = new Int2ObjectOpenHashMap<>();

	public FKFont(FKFontRef font) {
		this.fkFontRef = Objects.requireNonNull(font);
	}

	@Override
	public GlyphRef getGlyph(int codepoint) {
		// TODO: async rasterization
		int pageIdx = codepoint / PAGE_GLYPHS;
		var page = this.pages.computeIfAbsent(pageIdx, (IntFunction<Page>) this::rasterizePage);
		var glyphIdx = codepoint % PAGE_GLYPHS;
		return new GlyphRef(page.texture, (byte) glyphIdx, 0f, page.widths[glyphIdx]);
	}

	private Page rasterizePage(int pageIdx) {
		var buf = Arena.ofAuto().allocate(JAVA_INT, PAGE_SIZE);

		float[] widths = new float[PAGE_GLYPHS];

		int codepointBase = pageIdx * PAGE_GLYPHS;
		for (int i = 0; i < PAGE_GLYPHS; i++) {
			int codepoint = codepointBase + i;

			int x = i % PAGE_DIM_GLYPHS;
			int y = i / PAGE_DIM_GLYPHS;

			@SuppressWarnings("DataFlowIssue")
			var idOrNull = executeOn(() -> this.fkFontRef.glyphForCodepoint(codepoint), GLThread::submitToGLThread);
			//noinspection ConstantValue
			if (idOrNull == null) continue;
			int id = idOrNull;
			var rasterBounds = executeOn(() -> this.fkFontRef.rasterBounds(id, GLYPH_SIZE, RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
			widths[i] = rasterBounds.width() / (float) GLYPH_SIZE;
			var translationX = x * GLYPH_SIZE + rasterBounds.originX() * FACTORS.originX() + rasterBounds.width() * FACTORS.width();
//		var translationY = rasterBounds.originY() * this.factors.originY + rasterBounds.height() * this.factors.height;
			var translationY = y * GLYPH_SIZE + GLYPH_SIZE - (rasterBounds.originY() * FACTORS.originY() + rasterBounds.height() * FACTORS.height());
			executeOn(() -> this.fkFontRef.rasterizeCodepoint(buf,
					PAGE_DIM_SIZE,
					PAGE_DIM_SIZE,
					FKCanvasFormat.Rgba32,
					id,
					GLYPH_SIZE,
					translationX,
					translationY,
					RASTERIZATION_OPTIONS), GLThread::submitToGLThread);
		}
		unpremultiplyAlpha(buf);
		return new Page(new ImageDataLoadableTexture(new ImageData(PAGE_DIM_SIZE,
				PAGE_DIM_SIZE,
				ColorType.Rgba8,
				buf.asReadOnly())), widths);
	}

	private record Page(LoadableTexture texture, float[] widths) {}
}
